//  QUWidget.cpp version 1.5
//  yudit package - Unicode Editor for the X Window System (and Linux) 
//
//  Author: gsinai@iname.com (Gaspar Sinai)
//  GNU Copyright (C) 1997,1998,1999  Gaspar Sinai
// 
//  yudit version 1.5  Copyright(C) 30 November, 1999, Tokyo Japan  Gaspar Sinai
//  yudit version 1.4  Copyright(C) 25 November, 1999, Tokyo Japan  Gaspar Sinai
//  yudit version 1.3  Copyright(C)  5 April,    1999, Tokyo Japan  Gaspar Sinai
//  yudit version 1.2  Copyright(C) 10 December, 1998, Tokyo Japan  Gaspar Sinai
//  yudit version 1.1  Copyright(C) 23 August,   1998, Tokyo Japan  Gaspar Sinai
//  yudit version 1.0  Copyright(C) 17 May,      1998, Tokyo Japan  Gaspar Sinai
//  yudit version 0.99 Copyright(C)  4 April,    1998, Tokyo Japan  Gaspar Sinai
//  yudit version 0.97 Copyright(C)  4 February, 1998, Tokyo Japan  Gaspar Sinai
//  yudit version 0.95 Copyright(C) 10 January,  1998, Tokyo Japan  Gaspar Sinai
//  yudit version 0.94 Copyright(C) 17 December, 1997, Tokyo Japan  Gaspar Sinai
//  yudit version 0.9 Copyright (C)  8 December, 1997, Tokyo Japan  Gaspar Sinai
//  yutex version 0.8 Copyright (C)  5 November, 1997, Tokyo Japan  Gaspar Sinai
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//
// This is the main widget for th eexperimental DDD 
// Gaspar Sinai <gsinai@gol.com>
// Initial version 18 August 1997
//


#include <UXInput.h>
#include "QUWidget.h"
#include "QUXKinput.h"
#include <UCommon.h>
#include <qkeycode.h>
#include <qclipboard.h>
#include <UList.h>
#ifdef TrueColor
#undef TrueColor
#endif
#include <qapplication.h>

#define SCROLLBAR_SIZE  20
#define BLINK_TIME	500 // ms

static UList<QUWidget*> checkList;

#define EVENT_MASK (KeyPressMask | KeyReleaseMask | \
	ButtonPressMask | ButtonReleaseMask | \
	KeymapStateMask | \
	ButtonMotionMask | \
	EnterWindowMask | LeaveWindowMask | \
	FocusChangeMask | \
	ExposureMask | \
	StructureNotifyMask | SubstructureRedirectMask | \
	NoEventMask | PropertyChangeMask)


QUWidget::QUWidget (QWidget* parent, const char* font) : QFrame (parent)
{
	multiLine=UWidget::MULTILINE_SCROLL;
	xInputFont = 0;
	xInput = 0;
	xInputFontMap=0;

	setFrameStyle (Panel|Sunken);
	setLineWidth (2);
	twidget = new QUWidgetBase (this, "kyudit");
	twidget->uSetWidget (this->x11Display(), twidget->winId());
	twidget->uSetFont (font);
	XSelectInput (this->x11Display(), twidget->winId(),
		EVENT_MASK);
	hscrollBar = new QScrollBar (0, 0, // range
		1, 1, // line/page steps
		0,	// inital value
		QScrollBar::Horizontal,	// orientation
		this, "hscrollbar");
	vscrollBar = new QScrollBar (0, 0, 
		1, 1, // line/page steps
		0,	// inital value
		QScrollBar::Vertical,	// orientation
		this, "hscrollbar");
	hscrollBar->resize (this->width(), SCROLLBAR_SIZE);
	vscrollBar->resize (SCROLLBAR_SIZE, this->height());
	setMinimumSize (SCROLLBAR_SIZE, SCROLLBAR_SIZE);
	twidget->uSetFullSize ();
	connect (twidget, SIGNAL (setEVScroll (int, int, int, int)), this, SLOT (setVScroll (int, int, int, int)));
	connect (twidget, SIGNAL (setEHScroll (int, int, int, int)), this, SLOT (setHScroll (int, int, int, int)));
	connect (hscrollBar, SIGNAL (valueChanged (int)), twidget, 
		SLOT (hScroll (int)));
	connect (vscrollBar, SIGNAL (valueChanged (int)), twidget, 
		SLOT (vScroll (int)));
	connect( QApplication::clipboard(), SIGNAL(dataChanged()),
		twidget, SLOT(clipboardClear()) );
	twidget->uSetCursor (UCursor::INSERT);
	checkList.addItem (this, 0);
}

QUWidget::~QUWidget ()
{
	if (xInputFont!=0)
	{
		delete xInputFont;
		xInputFont = 0;
	}
	if (xInputFontMap!=0)
	{
		// We may use unusefontamp later.
		xInputFontMap = 0;
	}
	if (xInput!=0)
	{
		xInput->uEnd();
		uSetXInput (0);
		switch (xInputType)
		{
		case KinputType:
			delete ((QUXKinput*) xInput);
			break;
		default:
			delete xInput;
			break;
		}
		xInput=0;
	}
	if (hscrollBar) delete hscrollBar;
	if (vscrollBar) delete vscrollBar;
	if (twidget) delete twidget;
	checkList.deleteItem (this);
}

UString::UStatus
QUWidget::setConversion (const char *conversion)
{
	return twidget->uSetConverter (conversion);
}

UWidget::UStatus
QUWidget::setInput (const char *input)
{
	return twidget->uSetInput (input);
}

void
QUWidget::setTabSize (int newValue)
{
	twidget->uSetTabSize(newValue);
	twidget->uRefresh();
}

void
QUWidget::setBackground (const char* color)
{
	background.setNamedColor (color);
	twidget->uSetBackground((unsigned long)background.pixel());
	twidget->uRefresh();
}

void
QUWidget::setForeground (const char* color)
{
	foreground.setNamedColor (color);
	twidget->uSetForeground((unsigned long)foreground.pixel());
	twidget->uRefresh();
}

void
QUWidget::setCursorColor (const char* color)
{
	cursorColor.setNamedColor (color);
	twidget->uSetCursorColor((unsigned long)cursorColor.pixel());
	twidget->uRefresh();
}

void
QUWidget::setEditable (UWidget::UEditable editableIn)
{
	twidget->uSetEditable (editableIn);
}


void
QUWidget::appendText (const unsigned char *text, const int len)
{
	twidget->uAppendText (text, len, 0, 0);
}

unsigned char*
QUWidget::getLine (const int pos, int* len)
{
	
	return twidget->uGetLine (pos, len);
}


unsigned char*
QUWidget::getBuffer (int *len)
{
	
	return twidget->uGetBuffer (len);
}

void
QUWidget::uAddEventHadler (UXInput *xinput,
		UXInput::UInputEvent eventType)
{
	twidget->uEventHandler[eventType] = xinput;
}

void
QUWidget::uRemoveEventHadler (UXInput *xinput,
		UXInput::UInputEvent eventType)
{
	twidget->uEventHandler[eventType] = 0;
}

UWidget::UStatus
QUWidget::setFontName (const char* font)
{
	return twidget->uSetFont (font);
}

UWidget::UStatus
QUWidget::setFontSize (const int fontSize)
{
	return twidget->uSetFontSize (fontSize);
}

UWidget::UStatus
QUWidget::setFontAddStyle (const char* addStyle)
{
	return twidget->uSetFontAddStyle (addStyle);
}

UWidget::UStatus
QUWidget::setFontAvgWidth (const char* avgWidth)
{
	return twidget->uSetFontAvgWidth (avgWidth);
}

UWidget::UStatus
QUWidget::setFontWeight (const char* fontWeight)
{
	return twidget->uSetFontWeight (fontWeight);
}

UWidget::UStatus
QUWidget::setFontSlant (const char* fontSlant)
{
	return twidget->uSetFontSlant (fontSlant);
}

UWidget::UStatus
QUWidget::setFontSpacing (const char* fontSpacing)
{
	return twidget->uSetFontSpacing (fontSpacing);
}

void 
QUWidget::refresh()
{
	twidget->uRefresh();
}

void 
QUWidget::resizeEvent (QResizeEvent * event)
{
	twidget->setGeometry (
	  this->frameWidth(),
	  this->frameWidth(),
	  this->width() -frameWidth()*2 - ((vscrollBar==0) ? 0 : SCROLLBAR_SIZE), 
	  this->height() -frameWidth()*2 - ((hscrollBar==0) ? 0 : SCROLLBAR_SIZE));
	if (vscrollBar)
	{
		vscrollBar->setGeometry (
			this->width() - SCROLLBAR_SIZE - frameWidth() * 2, 
			this->frameWidth(),
			SCROLLBAR_SIZE, 
			this->height() - frameWidth()*2 - 
			((hscrollBar==0) ? 0 :SCROLLBAR_SIZE));
	}
	if (hscrollBar)
	{
		hscrollBar->setGeometry (
			this->frameWidth(), 
			this->height() - frameWidth()*2 -((vscrollBar==0) ? 0 : SCROLLBAR_SIZE),
			this->width() - frameWidth()*2 -((vscrollBar==0) ? 0 : SCROLLBAR_SIZE),
			SCROLLBAR_SIZE);
	}
	twidget->uSetFullSize ();
}

void
QUWidget::setVScroll (int value, int step, int page, int max)
{
	if (vscrollBar==0)
	{
		twidget->uVScroll (value);
		return;
	}
	if (max !=0) vscrollBar->setRange (0, (max-page > 0) ? max-page : 0);
	if (max !=0 || page !=0) vscrollBar->setSteps (step, page);
	if (value >=0) vscrollBar->setValue (value);
}

void
QUWidget::setHScroll (int value, int step, int page, int max)
{
	if (hscrollBar==0)
	{
		twidget->uHScroll (value);
		return;
	}
	if (max !=0) hscrollBar->setRange (0, (max-page > 0) ? max-page : 0);
	if (max !=0 || page !=0)  hscrollBar->setSteps (step, page);
	if (value >=0) hscrollBar->setValue (value);
}

const UCS2* const *
QUWidget::uGetTextBuffer()
{
	return twidget->uGetTextBuffer();
}

const int*
QUWidget::uGetTextSizes()
{
	return twidget->uGetTextSizes();
}

int
QUWidget::uGetTextSize()
{
	return twidget->uGetTextSize();
}

void
QUWidget::setMultiLine (UWidget::UMultiLine multiLineIn)
{
	twidget->uSetMultiLine (multiLineIn);
	switch (multiLineIn)
	{
	case UWidget::MULTILINE_SCROLL:
		if (hscrollBar==0)
		{
			hscrollBar = new QScrollBar (0, 0, // range
				1, 1, // line/page steps
				0,	// inital value
				QScrollBar::Horizontal,	// orientation
				this, "hscrollbar");
			hscrollBar->resize (this->width(), SCROLLBAR_SIZE);
			connect (hscrollBar, 
					SIGNAL (valueChanged (int)), twidget, 
				SLOT (hScroll (int)));
		}
		if (vscrollBar==0)
		{
			vscrollBar = new QScrollBar (0, 0, 
				1, 1, // line/page steps
				0,	// inital value
				QScrollBar::Vertical,	// orientation
				this, "hscrollbar");
			vscrollBar->resize (SCROLLBAR_SIZE, this->height());
			connect (vscrollBar, 
				SIGNAL (valueChanged (int)), twidget, 
				SLOT (vScroll (int)));
		}
		break;
	case UWidget::SINGLE_SCROLL:
		if (hscrollBar==0)
		{
			hscrollBar = new QScrollBar (0, 0, // range
				1, 1, // line/page steps
				0,	// inital value
				QScrollBar::Horizontal,	// orientation
				this, "hscrollbar");
			hscrollBar->resize (this->width(), SCROLLBAR_SIZE);
			connect (hscrollBar, 
					SIGNAL (valueChanged (int)), twidget, 
				SLOT (hScroll (int)));
		}
		if (vscrollBar!=0)
		{
			delete vscrollBar;
			vscrollBar=0;
		}
		break;
	default:
		if (hscrollBar!=0) delete hscrollBar;
		if (vscrollBar!=0) delete vscrollBar;
		hscrollBar=0;
		vscrollBar=0;
	}
	multiLine = multiLineIn;
	resize (this->width(), this->height());
}

int
QUWidget::uSetXInput (const char* name_, const char* imIn, 
	const char* inputStyleIn, const char* textType, 
	const char* uStringIn, const char* fontMap, double versionIn)
{
	UXInput::UInputStyle	inputStyle;

	if (xInputFont!=0)
	{
		delete xInputFont;
		xInputFont = 0;
	}
	if (xInputFontMap!=0)
	{
		// We may use unusefontamp later.
		xInputFontMap = 0;
	}
	if (xInput!=0)
	{
		xInput->uEnd();
		uSetXInput (0);
		switch (xInputType)
		{
		case KinputType:
			delete ((QUXKinput*) xInput);
			break;
		default:
			delete xInput;
			break;
		}
		xInput=0;
	}
	if (name_==0 || strcasecmp (name_, "None") ==0)
	{
		return 0;
	}
	xInputFontMap = UGetFontMap (fontMap);
	if (xInputFontMap==0)
	{
		cerr << "error: can not get fontmap '" << fontMap << "'.\n";
		return -1;
	}
	xInputFontMapName = fontMap;

	uSetCursor (UCursor::WAIT, 1);
	xInputFont = new UFontX11 (xInputFontMap,
		x11Display(), DefaultScreen (x11Display()));
	inputStyle = UXInput::UOFF;
	if (strcasecmp (inputStyleIn, "Root")==0)
	{
		inputStyle = UXInput::UROOT;
	}
	if (strcasecmp (name_, "Kinput2")==0)
	{
		xInputType = KinputType;
		xInput = new QUXKinput (this, getWidget());
		if (xInput->uSetup (imIn, inputStyle,
			textType, uStringIn, xInputFont, 
			(float) versionIn) != UXInput::OK)
		{
			delete ((QUXKinput*) xInput);
			xInput = 0;
			cerr << "error: can not set up xInput..";
			uSetCursor (UCursor::INSERT, 1);
			return -1;
		}
		uSetXInput (xInput);
		uSetCursor (UCursor::INSERT, 1);
		return 0;
	}
	uSetCursor (UCursor::INSERT, 1);
	return -1;
}

QUWidgetBase::QUWidgetBase (QWidget* parent, const char* name_) : QWidget (parent)
		, UWidget (name_, MULTILINE_SCROLL, EDITABLE)
{
	memset (uEventHandler, 0, UXInput::U_MAX_EVENT * sizeof (UXInput*));
	setFocusPolicy( StrongFocus );
	setActiveWindow ();
	setFocus ();
}

QUWidgetBase::~QUWidgetBase()
{
}

//
// In case you want to handle application events.
//
bool
QUWidget::uHandleXInputEvent (XEvent *ev)
{
	return FALSE;
}

bool
QUWidgetBase::x11Event (XEvent *ev)
{
	// let event proc happen.
	switch (ev->type)
	{
	case KeyPress:
		uHandleEvent (ev);
		return TRUE;
	case PropertyNotify:
		if (uEventHandler[UXInput::UPROPERTY_CHANGE]!=0)
		{
			uEventHandler[UXInput::UPROPERTY_CHANGE]->propertyChangeHandler (ev);
		}
		return TRUE;
	case ClientMessage:
		if (uEventHandler[UXInput::UNO_EVENT]!=0)
		{
			uEventHandler[UXInput::UNO_EVENT]->noEventHandler (ev);
		}
		return TRUE;
	}
	return FALSE;
}

/**
 * These routines will work with null terminated strings only
 */
void
QUWidgetBase::uClipboardPut (const unsigned char * text, const int len)
{
	disconnect( QApplication::clipboard(), SIGNAL(dataChanged()),
			this, 0);
	if (text==0)
	{
		QApplication::clipboard()->clear();
	}
	else
	{
		QApplication::clipboard()->setText ((const char*) text);
	}
	connect( QApplication::clipboard(), SIGNAL(dataChanged()),
		this, SLOT(clipboardClear()) );
}

/**
 * These routines will work with null terminated strings only
 */
unsigned char*
QUWidgetBase::uClipboardGet (int* len)
{
	char * 		chr;
	QString		string;

	string = QApplication::clipboard()->text();
	if (string.isEmpty()) return 0;
	chr = new char [string.length()+1];
	CHECKNULL (chr);
	strcpy (chr, (const char*) string);
	// our string is double null terminated
	chr[string.length()+1] = 0;
	*len = string.length();
	return (unsigned char*) chr;
}
void
QUWidgetBase::clipboardClear ()
{
	uClipboardClear();
}

void
QUWidgetBase::paintEvent (QPaintEvent * event)
{
	uRedraw (event->rect().x(), event->rect().y(), 
		event->rect().width(), event->rect().height());
}

void
QUWidgetBase::keyPressEvent (QKeyEvent *event)
{
	switch (event->key())
	{
		case Key_Left: 
			uProcKeyPressEvent (UK_LEFT, (const int)event->ascii());
			return; 
		case Key_Right: 
			uProcKeyPressEvent (UK_RIGHT, (const int)event->ascii());
			return; 
		case Key_Up: 
			uProcKeyPressEvent (UK_UP, (const int)event->ascii());
			return; 
		case Key_Down: 
			uProcKeyPressEvent (UK_DOWN, (const int)event->ascii());
			return; 
		case Key_Next:
			uProcKeyPressEvent (UK_NEXT, (const int)event->ascii());
			return; 
		case Key_Prior:
			uProcKeyPressEvent (UK_PRIOR, (const int)event->ascii());
			return; 
		case Key_Delete:
			uProcKeyPressEvent (UK_DELETE, (const int)event->ascii());
			return; 
		case Key_Backspace:
			uProcKeyPressEvent (UK_BACKSPACE, (const int)event->ascii());
			return; 
		case Key_Home:
			uProcKeyPressEvent (UK_HOME, (const int)event->ascii());
			return; 
		case Key_End:
			uProcKeyPressEvent (UK_END, (const int)event->ascii());
			return; 
		case Key_Return:
		case Key_Enter:
			uProcKeyPressEvent (UK_RETURN, (const int)event->ascii());
			return; 
		case Key_Escape:
			break;
		case Key_X:
			if (event->state() & ControlButton)
			{
				uProcKeyPressEvent (UK_XINPUT_TOGGLE, 
					(const int) event->ascii());
				return;
			}
			break;
	}
	uProcKeyPressEvent (UK_ASCII, (const int) event->ascii());
	return;
}

void
QUWidgetBase::mousePressEvent (QMouseEvent *event)
{
	if (event->button() == MidButton)
	{
		uProcMousePressEvent (UB_MIDDLE, event->pos().x(), event->pos().y());
	}
	else if (event->button() == LeftButton)
	{
		uProcMousePressEvent (UB_LEFT, event->pos().x(), event->pos().y());
	}
}
void
QUWidgetBase::mouseReleaseEvent (QMouseEvent *event)
{
	if (event->button() == MidButton)
	{
	}
	else if (event->button() == LeftButton)
	{
		uProcMouseReleaseEvent (UB_LEFT, event->pos().x(), event->pos().y());
	}
}

void
QUWidgetBase::mouseMoveEvent (QMouseEvent *event)
{
	if (event->state() & MidButton)
	{
	}
	else if (event->state() & LeftButton)
	{
		uProcMouseMotionEvent (UB_LEFT, event->pos().x(), event->pos().y());
	}
}

void
QUWidgetBase::timerEvent (QTimerEvent* event)
{
	uBlinkCursor();
}


void
QUWidgetBase::focusInEvent (QFocusEvent *e)
{
	killTimers ();
	startTimer (BLINK_TIME);
}

void
QUWidgetBase::focusOutEvent (QFocusEvent *e)
{
	killTimers ();
	uBlinkCursor(1);
}

void
QUWidgetBase::resizeEvent (QResizeEvent* event)
{
	uResize (this->width(), this->height());
}

void
QUWidgetBase::uSetVScroll (int value, int step, int page, int max)
{
	emit setEVScroll (value, step, page, max);
}

void
QUWidgetBase::uSetHScroll (int value, int step, int page, int max)
{
	emit setEHScroll (value, step, page, max);
}

void 
QUWidgetBase::vScroll (int value)
{
	uVScroll (value);
}

void 
QUWidgetBase::hScroll (int value)
{
	uHScroll (value);
}

void
QUWidgetBase::backgroundColorChange (const QColor &old)
{
	
	uSetBackground((unsigned long)old.pixel());
	uRefresh();
}

QUWidget*
QGetUWidgetItem (int i)
{
	return (QUWidget*) checkList.getItem(i);
}
